from math import degrees
from tkinter import font
import numpy as np
import pandas as pd
from astropy import units as u
from astropy.coordinates import SkyCoord
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
plt.rcParams.update({'font.size': 22})

def StructureFunctionAnalysis(path_to_datafile, bin_size,ColumnDensity,VelocityDispersion,SNR_limit=2):
    """StructureFunctionAnalysis

    This function implements Hildebrand's 2009 method to calculates the plane of the sky magnetic field using modified Davis-Chandrashekhar method. 
    The polarimetry data needs to be provied in a csv file with the following columns 'ra','dec','PA','ePA'
    The function also fits the line structure_function = np.sqrt(b**2 + l**2m**2)
    
    Args:
        path_to_datafile (string): relative path to the csv datafile with polarimetry data.
        bin_size (float): bin size to be used for calculating the structure function.
        ColumnDensity (float): column density of the target in number/cc units.
        VelocityDispersion (float): velocity dispersion value of the target in km/s
        SNR_limit (flaot): SNR ratio to be used to clean the data.
    Returns:
        b (float): b parameter from the strucutre function in radians
        structure_function (2D list): The 2D array consists of the strucutre fucntion and its corresponding (l parameter)
    """


    datafile = path_to_datafile
    l_bins_length = bin_size 
    ColDen = ColumnDensity
    VelDis = VelocityDispersion

    ####### importing the data
    data = pd.read_csv(datafile,delimiter=',')
    data = data[(data['P']/data['eP']) > SNR_limit]
    data = data[0<data['distance_parallax']/1000]
    ra = np.array(data['ra'])
    dec = np.array(data['dec'])
    PolAng = np.array(data['PA'])
    ePolAng = np.array(data['ePA'])

    ###### calculating the distance between two points on the celestial plane
    def Calc_l(ra1,dec1,ra2,dec2):
        c1 = SkyCoord(ra1,dec1,unit = 'deg')
        c2 = SkyCoord(ra2,dec2,unit = 'deg')
        sep = c1.separation(c2)
        return sep.arcminute

    distance_mat = np.ones((ra.shape[0],ra.shape[0]))*np.nan
    PolAngleSqr_diff = np.ones((ra.shape[0],ra.shape[0]))*np.nan
    ePolAngleSqr = np.ones((ra.shape[0],ra.shape[0]))*np.nan

    for i in range(ra.shape[0]):
        for j in range(i):
            if ((PolAng[i] - PolAng[j])>90):
                PolAngleSqr_diff[i,j] = (PolAng[i] - PolAng[j]-180)**2
                ePolAngleSqr[i,j] = ePolAng[i]**2+ePolAng[j]**2
                distance_mat[i,j] = Calc_l(ra[i],dec[i],ra[j],dec[j])

            elif ((PolAng[i] - PolAng[j])<-90):
                PolAngleSqr_diff[i,j] = (PolAng[i] - PolAng[j] + 180)**2
                ePolAngleSqr[i,j] = ePolAng[i]**2+ePolAng[j]**2
                distance_mat[i,j] = Calc_l(ra[i],dec[i],ra[j],dec[j])
            else:
                PolAngleSqr_diff[i,j] = (PolAng[i] - PolAng[j])**2
                ePolAngleSqr[i,j] = ePolAng[i]**2+ePolAng[j]**2
                distance_mat[i,j] = Calc_l(ra[i],dec[i],ra[j],dec[j])

    l_bins = np.arange(0,10,l_bins_length)
    l_bins_centre = (l_bins[:-1] + l_bins[1:])/2

    struc_func = []
    error_func = []
    error1_func = []
    count_array = []

    for k in range(l_bins.shape[0]-1):
        count = 0
        temp_PolAngle = []
        temp_ePolAngle = []
        temp_dist = []
        for i in range(PolAngleSqr_diff.shape[0]):
            for j in range(i):
                if l_bins[k]< distance_mat[i,j]<=l_bins[k+1]:
                    count += 1
                    temp_PolAngle.append(PolAngleSqr_diff[i,j])
                    temp_ePolAngle.append(ePolAngleSqr[i,j])
                    temp_dist.append(distance_mat[i,j])
                else:
                    pass
        if count != 0:
            temp_PolAngle = np.array(temp_PolAngle)
            temp_ePolAngle = np.array(temp_ePolAngle)
            std_PA = np.mean(temp_PolAngle) - np.mean(temp_ePolAngle)
            mean_PA = np.sqrt(std_PA)
            a = np.sqrt(temp_ePolAngle)
            err_PA1 = np.std(a)
            err_PA = err_PA1/np.sqrt(count)
            struc_func.append(mean_PA)
            error1_func.append(err_PA1)
            error_func.append(err_PA)
            count_array.append(count)
        else:
            struc_func.append(np.nan)
            error1_func.append(np.nan)
            error_func.append(np.nan)

    struc_func = np.array(struc_func)
    error_func = np.array(error_func)
    struc_func = np.sqrt(struc_func**2 - error_func**2)

    def fitfun(l, m, b):
        y = np.sqrt(b**2 + (m**2) * (l**2))
        return y

    df = pd.DataFrame({'l_bins': l_bins_centre,'SF':struc_func})
    df = df.dropna()
    popt, pcov = curve_fit(fitfun, df['l_bins'][0:3], df['SF'][0:3])
    line = np.linspace(0,10,100)
    fitted_line = fitfun(line,popt[0],popt[1])

    b = popt[1]
    b1 = popt[1]*np.pi/180
    perr = np.sqrt(np.diag(pcov))
    perr1 = perr*np.pi/180
    ratio = b1/(np.sqrt(2-(b1**2)))

    column_density = ColDen
    velocity_dispersion_FWHM = 2.35*VelDis
    Bmag_modCF = 9.3*(np.sqrt(2*column_density))*velocity_dispersion_FWHM*(1/b)

    print('perr',perr)
    print('Bmag',Bmag_modCF)
    print('ratio',ratio)
    print('Error in ratio', abs(b1/(np.sqrt(2-(b1**2))) - (b1 + perr1[1])/(np.sqrt(2-((b1+perr1[1])**2)))))
    print('b',b)
    print('Error in b',perr[1])
    print('b in radian',b1)
    print('Error in radian b',perr1[1])
    print('Error in ratio', abs(b1/(np.sqrt(2-(b1**2))) - (b1 - perr1[1])/(np.sqrt(2-((b1+perr1[1])**2)))))
    print('Error in ratio', abs(b1/(np.sqrt(2-(b1**2))) - (b1 + perr1[1])/(np.sqrt(2-((b1-perr1[1])**2)))))
    print('Error in ratio', abs(b1/(np.sqrt(2-(b1**2))) - (b1 - perr1[1])/(np.sqrt(2-((b1-perr1[1])**2)))))

    return b,[l_bins_centre ,struc_func]




from aplpy import FITSFigure  
import matplotlib.pyplot as plt 
import astropy.io.fits as fits
import astropy.units as u 
import numpy as np 
import os as os
import six

def PlotPolarizationAngles(Pol,PolAng,stokesI,stokesIerr,PolInten,background,vmin,vmax,colorbar_label,plot_title,steps=2,vector_color = 'cyan',vector_shadow = 'k',vector_length = 2,linewidth = 2,cmap='gray'):
	"""StructureFunctionAnalysis

	This function plots polarization angles on a given background
	The input needs to be an astropy.io.fits.PrimaryHDU

	Args:
		Pol (astropy.io.fits.PrimaryHDU): The Polarization fraction in astropy.io.fits.PrimaryHDU format.
		PolAng (astropy.io.fits.PrimaryHDU): The Polarization angles in astropy.io.fits.PrimaryHDU format.
		stokesI (astropy.io.fits.PrimaryHDU): The Stokes I file in astropy.io.fits.PrimaryHDU.
		stokesIerr (astropy.io.fits.PrimaryHDU): The error in Stokes I file in astropy.io.fits.PrimaryHDU.
		PolInten (astropy.io.fits.PrimaryHDU): The Polarization intensity in astropy.io.fits.PrimaryHDU format.
		background (astropy.io.fits.PrimaryHDU): The background on which the polarization angles needs to plotted in astropy.io.fits.PrimaryHDU format.
		vmin (float): The vmin value for the background in aplpy.show_colorscale
		vmax (float): The vmax value for the background in aplpy.show_colorscale
		colorbar_label (string): The label for colorbar
		plot_title (string): The title for the figure
		steps (int): Steps of the vectors plotted. Default is 2 plotting every 2nd vector, choosing 1 means plotting every vector
		vector_color (string): Color of the vectors. Default is cyan.
		vector_shadow (string): Color of the vectors. Default is black
		vector_length (float): The length of the vector. Default is 2
		linewidth (float): The width of the vector. Default is 2
		cmap (string): Color scheme for the plot	
	Returns:
		Plots the vectors on the background.
	"""
	p    = Pol
	pa   = PolAng
	stkI = stokesI
	stkIerr = stokesIerr
	pi = PolInten
	title = plot_title
	pxscale = stkI.header['CDELT2']*3600
	stkI.data /= pxscale**2
	pi.data /= pxscale**2 
	stkIerr.data /= pxscale**2
	title_size = 16
	tick_colorbar = 15
	scalevec = vector_length #1px = scalevec * 1% pol 
	vec_legend = 5.0

	#### SCRIPT
	fig = plt.figure(figsize=(13,10))
	scuba = FITSFigure(background,figure=fig)

	#colorscale
	scuba.show_colorscale(cmap=cmap,vmin = vmin,vmax = vmax)



	#colorbar
	scuba.add_colorbar(location='right', width=0.2, pad=0.15, ticks=None,axis_label_text= colorbar_label)
	scuba.colorbar.set_font(size=tick_colorbar)
	scuba.set_title(title,fontsize=title_size)

	scuba.show_vectors(p,pa,scale=scalevec,step=steps,color=vector_shadow,linewidth=linewidth*1.5)
	scuba.show_vectors(p,pa,scale=scalevec,step=steps,color=vector_color,linewidth=linewidth)

	vecscale = scalevec * pxscale/3600
	scuba.add_scalebar(vec_legend*vecscale,r'$p_{frac}$ ='+np.str(vec_legend),corner='bottom right',frame=True,color='black')
	plt.show()




from aplpy import FITSFigure  
import matplotlib.pyplot as plt 
import astropy.io.fits as fits
import astropy.units as u 
import numpy as np 
import os as os
import six
import pandas as pd
plt.rcParams.update({'font.size': 18})

def PlanckPolarization(path_to_stokesI,path_to_stokesQ,path_to_stokesU,path_to_background,angle_offset,vmin,vmax,line_width,vector_scale,steps = 2,vector_color= 'cyan',cmap = 'gray'):
    """PlanckPolarization

    This function plots the magnetic fields from Planck Polarization vecotrs. The function requires the Planck polarization stokes I,Q and U map in celestial coordiantes (fk5).
    The function also requires a variable angle_offset which is the angle between the galactic and celestial north at the position of the target
    
    Args:
        path_to_stokesI (string): relative path to the fits file of the Planck stokes I file in celestial coordinates.
        path_to_stokesQ (string): relative path to the fits file of the Planck stokes Q file in celestial coordinates.
        path_to_stokesU (string): relative path to the fits file of the Planck stokes U file in celestial coordinates.
        path_to_background (string): relative path to the fits file of the background in celestial coordinates.
        angle_offset (float): The angle between the between the galactic and celestial north at the position of the target
        vmin = This parameter controls the minimum cutoff of the background for plotting
        vmax = This parameter controls the minimum cutoff of the background for plotting
        line_width (float): This parameter controls the width of the vector
        steps (integer): This paramter controls the vectors skipped. If steps is 1 it plots every vector, default is 2 which plots every second vector
        vector_scale (flaot): This parameter controls the length of the vector
        vector_color (string): Color of the vectors. Default cyan
        cmap (string): Colormap of the background. Default gray

    Returns:
        Plots the magnetic fields from Planck Polarization vectors
    """

    def get_data(stokesI,stokesQ,stokesU,angle_offset):
        stokesIfile = fits.open(stokesI)
        stokesQfile = fits.open(stokesQ)
        stokesUfile = fits.open(stokesU)
        plank_polarization = stokesIfile[0].copy()
        plank_theta_mod = stokesIfile[0].copy()
        plank_polarization.data = np.sqrt(stokesQfile[0].data*stokesQfile[0].data + stokesUfile[0].data*stokesUfile[0].data)/stokesIfile[0].data
        plank_theta_mod.data = (180/np.pi)*0.5*np.arctan2(-1*stokesUfile[0].data,stokesQfile[0].data) + 90 + angle_offset

        return plank_polarization,plank_theta_mod
    p,pa = get_data(path_to_stokesI,path_to_stokesQ,path_to_stokesU,angle_offset)
    Background = fits.open(path_to_background)[0]

    ###Figure
    scalevec = vector_scale #1px = scalevec * 1% pol 
    p.data = p.data/p.data
    fig = plt.figure(figsize=(13,10))
    gc = FITSFigure(Background,figure=fig)
    gc.show_colorscale(cmap=cmap,vmin=vmin,vmax =vmax)
    gc.show_vectors(p,pa,scale=scalevec,step=steps,color=vector_color,linewidth=line_width)
    gc.show_vectors(p,pa,scale=scalevec,step=steps,color='k',linewidth=line_width*1.1)
    plt.show()





import numpy as np
import matplotlib.pyplot as plt
import matplotlib
import aplpy
from astropy.wcs import WCS
from astroquery.gaia import Gaia 
import pandas as pd
from astropy import units as u
from astropy.io import fits
from astropy.table import Table
from astropy.coordinates import SkyCoord
from scipy.optimize import curve_fit
from astropy.stats import bootstrap
import matplotlib.ticker as ticker
plt.rcParams.update({'font.size': 18})
plt.rcParams['xtick.labelsize']=16
plt.rcParams['ytick.labelsize']=16

def paramters_search(ra_coord,dec_coord):
    """parameters_search

    Given RA and DEC coordinates this program finds the parallax and distance using GAIA data release 3
    
    Args:
        ra_coord (float): RA coordinate at which to search 
        dec_coord (float): DEC coordinate at which to search 
    Returns:
        parallax: (float): Gaia parallax of the object at the given RA and DEC
        dist: (float): Hipparcos distance to the object at the given RA and DEC
        search_width (float): Search area in which object was found
    """
    Gaia.MAIN_GAIA_TABLE = "gaiadr3.gaia_source" # select Gaia release 

    search_width = 0 #initialising the search width
    flag = 0 
    while flag == 0: # initialize the while loop till we find an object within search width
        search_width += 0.25
        coord = SkyCoord(ra=ra_coord, dec=dec_coord, unit=(u.degree, u.degree),frame='icrs')
        width = u.Quantity(search_width, u.arcsecond)
        height = u.Quantity(search_width, u.arcsecond)
        results = Gaia.query_object(coordinate=coord, width=width, height=height)
        flag  = results['parallax'].shape[0]
        if abs(search_width - 30*0.25)< 0.2:  # condition so that the search width doesn't increase 0.25 degrees
            return -1,-1,-1 # returns -1 values as parallax and distances can't be negative 
    print(search_width) 
    return results['parallax'][0],results['dist'][0],search_width



def generate_RA_DEC_mesh(hdu):
    """generate_RA_DEC_mesh

    Given the fits file HDU the function generates the RA and DEC grids
    
    Args:
        hdu (astropy fits): RA coordinate at which to search 
    Returns:
        DEC_grid (2D array): DEC grid for the given fits file
        RA_grid (2D array): RA grid for the given fits file
    """
    if 'CDELT1' in hdu.header:
        RA_delt = hdu.header['CDELT1']
        DEC_delt = hdu.header['CDELT2']

    if 'CD1_1' in hdu.header:
        RA_delt = hdu.header['CD1_1']
        DEC_delt = hdu.header['CD2_2']

    RA_ref = (hdu.header['CRPIX1'])
    DEC_ref = (hdu.header['CRPIX2'])
    RA_ref_value = hdu.header['CRVAL1']
    DEC_ref_value = hdu.header['CRVAL2']
    RA_axis_len = hdu.header['NAXIS1']
    DEC_axis_len = hdu.header['NAXIS2']

    RA_axis = np.arange(1,RA_axis_len+1)
    DEC_axis = np.arange(1,DEC_axis_len+1)
    DEC_axis_modified = np.arange(1,RA_axis_len+1)
    
    DEC_array = (DEC_axis - DEC_axis_len/2)*DEC_delt + DEC_ref_value
    DEC_array_modified = (DEC_axis_modified - RA_axis_len/2)*DEC_delt + DEC_ref_value
    RA_array = RA_ref_value-(RA_axis - RA_axis_len/2)*(RA_delt*(-1)/np.cos(DEC_array_modified*0.01745))

    # #making a meshgrid from the arrays
    DEC_grid,RA_grid = np.meshgrid(DEC_array,RA_array , sparse=False, indexing='ij')
    return DEC_grid,RA_grid



def Calc_l(ra1,dec1,ra2,dec2):
    """Calc_l

    The function calculates the distance between two points on the celestial plane
    
    Args:
        ra1 (float): RA coordinate of first position in degrees
        dec1 (float): DEC coordinate of first position in degrees
        ra2 (float): RA coordinate of second position in degrees
        dec2 (float): DEC coordinate of second position in degrees
    Returns:
        seperation (float): returns the seperation in arcminutes between the two points on the celestial sphere

    """
    c1 = SkyCoord(ra1,dec1,unit = 'deg')
    c2 = SkyCoord(ra2,dec2,unit = 'deg')
    sep = c1.separation(c2)
    return sep.arcminute


def wrapper(Angle_grid):
    """wrapper

    The function keeps the polarization angles in a array between -90 and 90
    
    Args:
        Angle_grid (N-D numpy array): N dimensional array consisting of polarization angles
    Returns:
        Angle_grid (N-D numpy array): N dimensional array consisting of modified polarization angles
    """
    while ((np.nanmax(Angle_grid)>90) or (np.nanmin(Angle_grid)<-90)):
        Angle_selector =Angle_grid>90
        Angle_grid[Angle_selector] = Angle_grid[Angle_selector] - 180
        Angle_selector = Angle_grid<-90
        Angle_grid[Angle_selector] = Angle_grid[Angle_selector] + 180
    return Angle_grid

    
def remove_nan(array1,array2):
    """remove_nan

    The function removes the NAN values in two arrays to create modified arrays of same size. Useful for removing NAN values 
    in two arrays with one being the function of other
    
    Args:
        array1 (numpy array): first array to be filtered
        array2 (numpy array): Second array to be filtered
    Returns:
        array1_fil (numpy array): first array filtered
        array2_fil (numpy array): Second array filtered
    """
    selector = ~np.isnan(array1)
    array1_fil = array1[selector]
    array2_fil = array2[selector]

    selector = ~np.isnan(array2_fil)
    array1_fil = array1_fil[selector]
    array2_fil = array2_fil[selector]
    return array1_fil,array2_fil

def quality_cuts(stokesI,stokesIerr,Pol,Polerr,SNRp_cut,p_cut,SNRi_cut):
    """quality_cuts

    The function does the signal to noise ratio qualtity cuts for polarizaiton fraction.

    Args:
		stokesI (astropy.io.fits.PrimaryHDU): The Stokes I file in astropy.io.fits.PrimaryHDU.
		stokesIerr (astropy.io.fits.PrimaryHDU): The error in Stokes I file in astropy.io.fits.PrimaryHDU.
		Pol (astropy.io.fits.PrimaryHDU): The Polarization fraction in astropy.io.fits.PrimaryHDU format.
		Polerr (astropy.io.fits.PrimaryHDU): The error in Polarization fraction in astropy.io.fits.PrimaryHDU format.
		SNRp_cut (float): Signal to noise ratio cutoff for Polarizaiton fraction.
		p_cut (float): Maximum value cutoff for Polarizaiton fraction.
        SNRi_cut (float): Signal to noise ratio cutoff for Intensity.
    Returns:
	    Pol (astropy.io.fits.PrimaryHDU): The filtered Polarization fraction in astropy.io.fits.PrimaryHDU format.
    """
	#snr in P
    SNRp = Pol.data/Polerr.data
    mask_snrp = np.where(SNRp < SNRp_cut)
    Pol.data[mask_snrp] = np.nan
    #p_cut
    maskp = np.where(Pol.data > p_cut)
    Pol.data[maskp] = np.nan
    #snr in P
    SNRi = stokesI.data/stokesIerr.data
    mask_snri = np.where(SNRi < SNRi_cut)
    Pol.data[mask_snri] = np.nan
    return Pol


import numpy as np
import matplotlib.pyplot as plt
import matplotlib
import aplpy
import pandas as pd
from astropy.io import fits
from astropy.table import Table
from astropy.coordinates import SkyCoord
from scipy.optimize import curve_fit
from astropy.stats import bootstrap
import matplotlib.ticker as ticker
plt.rcParams.update({'font.size': 18})
plt.rcParams['xtick.labelsize']=16
plt.rcParams['ytick.labelsize']=16
from Utilities import generate_RA_DEC_mesh
from Utilities import Calc_l
from Utilities import wrapper


def AngularDispersion(stokesQ,stokesU,PolAngerr,lag = 0.5):    
    """StructureFunctionAnalysis

    This function implements Planck Collaboration 2016 method to calculates the angular dispersion of the region
    The input needs to be an astropy.io.fits.PrimaryHDU
    
    Args:
        stokesQ (astropy.io.fits.PrimaryHDU): The Stokes Q file in astropy.io.fits.PrimaryHDU.
        stokesU (astropy.io.fits.PrimaryHDU): The Stokes U file in astropy.io.fits.PrimaryHDU.
        PolAngerr (astropy.io.fits.PrimaryHDU): The error in Polarization angles in astropy.io.fits.PrimaryHDU format.
        lag (float): The lag also referred to as delta is specifies the area under which the angular dispersion is calculated, For more information please refer to the paper mentioned above.
        
    Returns:
        S_map_deb (2D numpy array): returns the debiased angular dispersion map of the region.
    """

    ############## generating the RA and DEC mesh
    DEC_grid,RA_grid = generate_RA_DEC_mesh(stokesQ)
    seperation = stokesQ.copy()
    set_delta = lag   # in arcminute
    S_map = stokesQ.copy()
    sigma_S_map = stokesQ.copy()

    for i in range(RA_grid.shape[0]):
        for j in range(RA_grid.shape[1]):
            ##### seperation filter
            seperation.data = Calc_l(RA_grid[i,j],DEC_grid[i,j],RA_grid,DEC_grid)
            seperation_selector = (seperation.data<0.5*set_delta)
            seperation.data[seperation_selector] = np.nan
            seperation_selector = (seperation.data>1.5*set_delta)
            seperation.data[seperation_selector] = np.nan
            seperation_selector = (seperation.data >0)

            ##### making the dispersion map
            tempa = stokesQ.data*stokesU.data[i,j] - stokesQ.data[i,j]*stokesU.data
            tempb = stokesQ.data*stokesQ.data[i,j] + stokesU.data*stokesU.data[i,j]
            AngleDiff_v2 = 0.5 * (180/np.pi)*np.arctan2(tempa,tempb)
            S = np.nanmean(AngleDiff_v2[seperation_selector]**2)**0.5
            S_map.data[i,j] = S

            ##### making the dispersion error map
            sigma_S = np.nanmean(PolAngerr.data[seperation_selector]**2)**0.5
            sigma_S_map.data[i,j] = sigma_S
    S_map_deb = S_map.copy()
    S_map_deb.data = np.sqrt(S_map.data**2 - sigma_S_map.data**2)
    return S_map_deb